@@ -1,13 +1,14 @@
 # Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
 # For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt
 
-"""Core control stuff for coverage.py."""
+"""Central control stuff for coverage.py."""
 
 from __future__ import annotations
 
 import atexit
 import collections
 import contextlib
+import functools
 import os
 import os.path
 import platform
@@ -20,16 +21,18 @@ import warnings
 from types import FrameType
 from typing import (
     cast,
-    Any, Callable, Dict, IO, Iterable, Iterator, List, Optional, Tuple, Union,
+    Any, Callable, IO, Iterable, Iterator, List,
 )
 
 from coverage import env
 from coverage.annotate import AnnotateReporter
 from coverage.collector import Collector, HAS_CTRACER
-from coverage.config import read_coverage_config
+from coverage.config import CoverageConfig, read_coverage_config
 from coverage.context import should_start_context_test_function, combine_context_switchers
 from coverage.data import CoverageData, combine_parallel_data
-from coverage.debug import DebugControl, short_stack, write_formatted_info
+from coverage.debug import (
+    DebugControl, NoDebugging, short_stack, write_formatted_info, relevant_environment_display,
+)
 from coverage.disposition import disposition_debug_msg
 from coverage.exceptions import ConfigError, CoverageException, CoverageWarning, PluginError
 from coverage.files import PathAliases, abs_file, relative_filename, set_relative_directory
@@ -37,22 +40,21 @@ from coverage.html import HtmlReporter
 from coverage.inorout import InOrOut
 from coverage.jsonreport import JsonReporter
 from coverage.lcovreport import LcovReporter
-from coverage.misc import bool_or_none, join_regex, human_sorted
+from coverage.misc import bool_or_none, join_regex
 from coverage.misc import DefaultValue, ensure_dir_for_file, isolate_module
 from coverage.multiproc import patch_multiprocessing
 from coverage.plugin import FileReporter
 from coverage.plugin_support import Plugins
 from coverage.python import PythonFileReporter
-from coverage.report import render_report
-from coverage.results import Analysis
-from coverage.summary import SummaryReporter
+from coverage.report import SummaryReporter
+from coverage.report_core import render_report
+from coverage.results import Analysis, analysis_from_file_reporter
 from coverage.types import (
-    TConfigurable, TConfigSectionIn, TConfigValueIn, TConfigValueOut,
+    FilePath, TConfigurable, TConfigSectionIn, TConfigValueIn, TConfigValueOut,
     TFileDisposition, TLineNo, TMorf,
 )
 from coverage.xmlreport import XmlReporter
 
-
 os = isolate_module(os)
 
 @contextlib.contextmanager
@@ -85,7 +87,14 @@ class Coverage(TConfigurable):
         cov.start()
         #.. call your code ..
         cov.stop()
-        cov.html_report(directory='covhtml')
+        cov.html_report(directory="covhtml")
+
+    A context manager is available to do the same thing::
+
+        cov = Coverage()
+        with cov.collect():
+            #.. call your code ..
+        cov.html_report(directory="covhtml")
 
     Note: in keeping with Python custom, names starting with underscore are
     not part of the public API. They might stop working at any point.  Please
@@ -96,10 +105,10 @@ class Coverage(TConfigurable):
     """
 
     # The stack of started Coverage instances.
-    _instances: List[Coverage] = []
+    _instances: list[Coverage] = []
 
     @classmethod
-    def current(cls) -> Optional[Coverage]:
+    def current(cls) -> Coverage | None:
         """Get the latest started `Coverage` instance, if any.
 
         Returns: a `Coverage` instance, or None.
@@ -114,21 +123,21 @@ class Coverage(TConfigurable):
 
     def __init__(                       # pylint: disable=too-many-arguments
         self,
-        data_file: Optional[Union[str, DefaultValue]] = DEFAULT_DATAFILE,
-        data_suffix: Optional[Union[str, bool]] = None,
-        cover_pylib: Optional[bool] = None,
+        data_file: FilePath | DefaultValue | None = DEFAULT_DATAFILE,
+        data_suffix: str | bool | None = None,
+        cover_pylib: bool | None = None,
         auto_data: bool = False,
-        timid: Optional[bool] = None,
-        branch: Optional[bool] = None,
-        config_file: Union[str, bool] = True,
-        source: Optional[Iterable[str]] = None,
-        source_pkgs: Optional[Iterable[str]] = None,
-        omit: Optional[Union[str, Iterable[str]]] = None,
-        include: Optional[Union[str, Iterable[str]]] = None,
-        debug: Optional[Iterable[str]] = None,
-        concurrency: Optional[Union[str, Iterable[str]]] = None,
+        timid: bool | None = None,
+        branch: bool | None = None,
+        config_file: FilePath | bool = True,
+        source: Iterable[str] | None = None,
+        source_pkgs: Iterable[str] | None = None,
+        omit: str | Iterable[str] | None = None,
+        include: str | Iterable[str] | None = None,
+        debug: Iterable[str] | None = None,
+        concurrency: str | Iterable[str] | None = None,
         check_preimported: bool = False,
-        context: Optional[str] = None,
+        context: str | None = None,
         messages: bool = False,
     ) -> None:
         """
@@ -219,14 +228,20 @@ class Coverage(TConfigurable):
             The `messages` parameter.
 
         """
+        # Start self.config as a usable default configuration. It will soon be
+        # replaced with the real configuration.
+        self.config = CoverageConfig()
+
         # data_file=None means no disk file at all. data_file missing means
         # use the value from the config file.
         self._no_disk = data_file is None
         if isinstance(data_file, DefaultValue):
             data_file = None
+        if data_file is not None:
+            data_file = os.fspath(data_file)
 
         # This is injectable by tests.
-        self._debug_file: Optional[IO[str]] = None
+        self._debug_file: IO[str] | None = None
 
         self._auto_load = self._auto_save = auto_data
         self._data_suffix_specified = data_suffix
@@ -235,23 +250,25 @@ class Coverage(TConfigurable):
         self._warn_no_data = True
         self._warn_unimported_source = True
         self._warn_preimported_source = check_preimported
-        self._no_warn_slugs: List[str] = []
+        self._no_warn_slugs: list[str] = []
         self._messages = messages
 
         # A record of all the warnings that have been issued.
-        self._warnings: List[str] = []
+        self._warnings: list[str] = []
 
-        # Other instance attributes, set later.
-        self._debug: DebugControl
-        self._plugins: Plugins
-        self._inorout: InOrOut
-        self._data: CoverageData
-        self._collector: Collector
-        self._file_mapper: Callable[[str], str]
+        # Other instance attributes, set with placebos or placeholders.
+        # More useful objects will be created later.
+        self._debug: DebugControl = NoDebugging()
+        self._inorout: InOrOut | None = None
+        self._plugins: Plugins = Plugins()
+        self._data: CoverageData | None = None
+        self._collector: Collector | None = None
+        self._metacov = False
 
+        self._file_mapper: Callable[[str], str] = abs_file
         self._data_suffix = self._run_suffix = None
-        self._exclude_re: Dict[str, str] = {}
-        self._old_sigterm: Optional[Callable[[int, Optional[FrameType]], Any]] = None
+        self._exclude_re: dict[str, str] = {}
+        self._old_sigterm: Callable[[int, FrameType | None], Any] | None = None
 
         # State machine variables:
         # Have we initialized everything?
@@ -263,6 +280,8 @@ class Coverage(TConfigurable):
         self._should_write_debug = True
 
         # Build our configuration from a number of sources.
+        if not isinstance(config_file, bool):
+            config_file = os.fspath(config_file)
         self.config = read_coverage_config(
             config_file=config_file,
             warn=self._warn,
@@ -303,10 +322,10 @@ class Coverage(TConfigurable):
 
         self._inited = True
 
-        # Create and configure the debugging controller. COVERAGE_DEBUG_FILE
-        # is an environment variable, the name of a file to append debug logs
-        # to.
-        self._debug = DebugControl(self.config.debug, self._debug_file)
+        # Create and configure the debugging controller.
+        self._debug = DebugControl(self.config.debug, self._debug_file, self.config.debug_file)
+        if self._debug.should("process"):
+            self._debug.write("Coverage._init")
 
         if "multiprocessing" in (self.config.concurrency or ()):
             # Multi-processing uses parallel for the subprocesses, so also use
@@ -317,7 +336,8 @@ class Coverage(TConfigurable):
         self._exclude_re = {}
 
         set_relative_directory()
-        self._file_mapper = relative_filename if self.config.relative_files else abs_file
+        if self.config.relative_files:
+            self._file_mapper = relative_filename
 
         # Load plugins
         self._plugins = Plugins.load_plugins(self.config.plugins, self.config, self._debug)
@@ -336,10 +356,10 @@ class Coverage(TConfigurable):
             self._should_write_debug = False
             self._write_startup_debug()
 
-        # '[run] _crash' will raise an exception if the value is close by in
+        # "[run] _crash" will raise an exception if the value is close by in
         # the call stack, for testing error handling.
-        if self.config._crash and self.config._crash in short_stack(limit=4):
-            raise Exception(f"Crashing because called by {self.config._crash}")
+        if self.config._crash and self.config._crash in short_stack():
+            raise RuntimeError(f"Crashing because called by {self.config._crash}")
 
     def _write_startup_debug(self) -> None:
         """Write out debug info at startup if needed."""
@@ -371,8 +391,9 @@ class Coverage(TConfigurable):
         Calls `_should_trace_internal`, and returns the FileDisposition.
 
         """
+        assert self._inorout is not None
         disp = self._inorout.should_trace(filename, frame)
-        if self._debug.should('trace'):
+        if self._debug.should("trace"):
             self._debug.write(disposition_debug_msg(disp))
         return disp
 
@@ -382,8 +403,9 @@ class Coverage(TConfigurable):
         Returns a boolean: True if the file should be traced, False if not.
 
         """
+        assert self._inorout is not None
         reason = self._inorout.check_include_omit_etc(filename, frame)
-        if self._debug.should('trace'):
+        if self._debug.should("trace"):
             if not reason:
                 msg = f"Including {filename!r}"
             else:
@@ -392,7 +414,7 @@ class Coverage(TConfigurable):
 
         return not reason
 
-    def _warn(self, msg: str, slug: Optional[str] = None, once: bool = False) -> None:
+    def _warn(self, msg: str, slug: str | None = None, once: bool = False) -> None:
         """Use `msg` as a warning.
 
         For warning suppression, use `slug` as the shorthand.
@@ -402,9 +424,7 @@ class Coverage(TConfigurable):
 
         """
         if not self._no_warn_slugs:
-            # _warn() can be called before self.config is set in __init__...
-            if hasattr(self, "config"):
-                self._no_warn_slugs = list(self.config.disable_warnings)
+            self._no_warn_slugs = list(self.config.disable_warnings)
 
         if slug in self._no_warn_slugs:
             # Don't issue the warning
@@ -413,7 +433,7 @@ class Coverage(TConfigurable):
         self._warnings.append(msg)
         if slug:
             msg = f"{msg} ({slug})"
-        if hasattr(self, "_debug") and self._debug.should('pid'):
+        if self._debug.should("pid"):
             msg = f"[{os.getpid()}] {msg}"
         warnings.warn(msg, category=CoverageWarning, stacklevel=2)
 
@@ -426,7 +446,7 @@ class Coverage(TConfigurable):
         if self._messages:
             print(msg)
 
-    def get_option(self, option_name: str) -> Optional[TConfigValueOut]:
+    def get_option(self, option_name: str) -> TConfigValueOut | None:
         """Get an option from the configuration.
 
         `option_name` is a colon-separated string indicating the section and
@@ -444,7 +464,7 @@ class Coverage(TConfigurable):
         """
         return self.config.get_option(option_name)
 
-    def set_option(self, option_name: str, value: Union[TConfigValueIn, TConfigSectionIn]) -> None:
+    def set_option(self, option_name: str, value: TConfigValueIn | TConfigSectionIn) -> None:
         """Set an option in the configuration.
 
         `option_name` is a colon-separated string indicating the section and
@@ -479,19 +499,20 @@ class Coverage(TConfigurable):
     def load(self) -> None:
         """Load previously-collected coverage data from the data file."""
         self._init()
-        if hasattr(self, "_collector"):
+        if self._collector is not None:
             self._collector.reset()
         should_skip = self.config.parallel and not os.path.exists(self.config.data_file)
         if not should_skip:
             self._init_data(suffix=None)
         self._post_init()
         if not should_skip:
+            assert self._data is not None
             self._data.read()
 
     def _init_for_start(self) -> None:
         """Initialization for start()"""
         # Construct the collector.
-        concurrency: List[str] = self.config.concurrency or []
+        concurrency: list[str] = self.config.concurrency or []
         if "multiprocessing" in concurrency:
             if self.config.config_file is None:
                 raise ConfigError("multiprocessing requires a configuration file")
@@ -520,6 +541,7 @@ class Coverage(TConfigurable):
             branch=self.config.branch,
             warn=self._warn,
             concurrency=concurrency,
+            metacov=self._metacov,
         )
 
         suffix = self._data_suffix_specified
@@ -537,6 +559,7 @@ class Coverage(TConfigurable):
 
         self._init_data(suffix)
 
+        assert self._data is not None
         self._collector.use_data(self._data, self.config.context)
 
         # Early warning if we aren't going to be able to support plugins.
@@ -548,7 +571,7 @@ class Coverage(TConfigurable):
                             for plugin in self._plugins.file_tracers
                     ),
                     self._collector.tracer_name(),
-                )
+                ),
             )
             for plugin in self._plugins.file_tracers:
                 plugin._coverage_enabled = False
@@ -557,7 +580,7 @@ class Coverage(TConfigurable):
         self._inorout = InOrOut(
             config=self.config,
             warn=self._warn,
-            debug=(self._debug if self._debug.should('trace') else None),
+            debug=(self._debug if self._debug.should("trace") else None),
             include_namespace_packages=self.config.include_namespace_packages,
         )
         self._inorout.plugins = self._plugins
@@ -578,9 +601,9 @@ class Coverage(TConfigurable):
                     signal.SIGTERM, self._on_sigterm,
                 )
 
-    def _init_data(self, suffix: Optional[Union[str, bool]]) -> None:
+    def _init_data(self, suffix: str | bool | None) -> None:
         """Create a data file if we don't have one yet."""
-        if not hasattr(self, "_data"):
+        if self._data is None:
             # Create the data file.  We do this at construction time so that the
             # data file will be written into the directory where the process
             # started rather than wherever the process eventually chdir'd to.
@@ -596,13 +619,16 @@ class Coverage(TConfigurable):
     def start(self) -> None:
         """Start measuring code coverage.
 
-        Coverage measurement only occurs in functions called after
+        Coverage measurement is only collected in functions called after
         :meth:`start` is invoked.  Statements in the same scope as
         :meth:`start` won't be measured.
 
         Once you invoke :meth:`start`, you must also call :meth:`stop`
         eventually, or your process might not shut down cleanly.
 
+        The :meth:`collect` method is a context manager to handle both
+        starting and stopping collection.
+
         """
         self._init()
         if not self._inited_for_start:
@@ -610,6 +636,9 @@ class Coverage(TConfigurable):
             self._init_for_start()
         self._post_init()
 
+        assert self._collector is not None
+        assert self._inorout is not None
+
         # Issue warnings for possible problems.
         self._inorout.warn_conflicting_settings()
 
@@ -631,19 +660,33 @@ class Coverage(TConfigurable):
             if self._instances[-1] is self:
                 self._instances.pop()
         if self._started:
+            assert self._collector is not None
             self._collector.stop()
         self._started = False
 
+    @contextlib.contextmanager
+    def collect(self) -> Iterator[None]:
+        """A context manager to start/stop coverage measurement collection.
+
+        .. versionadded:: 7.3
+
+        """
+        self.start()
+        try:
+            yield
+        finally:
+            self.stop()
+
     def _atexit(self, event: str = "atexit") -> None:
         """Clean up on process shutdown."""
         if self._debug.should("process"):
             self._debug.write(f"{event}: pid: {os.getpid()}, instance: {self!r}")
         if self._started:
             self.stop()
-        if self._auto_save:
+        if self._auto_save or event == "sigterm":
             self.save()
 
-    def _on_sigterm(self, signum_unused: int, frame_unused: Optional[FrameType]) -> None:
+    def _on_sigterm(self, signum_unused: int, frame_unused: FrameType | None) -> None:
         """A handler for signal.SIGTERM."""
         self._atexit("sigterm")
         # Statements after here won't be seen by metacov because we just wrote
@@ -660,11 +703,12 @@ class Coverage(TConfigurable):
         """
         self._init()
         self._post_init()
-        if hasattr(self, "_collector"):
+        if self._collector is not None:
             self._collector.reset()
         self._init_data(suffix=None)
+        assert self._data is not None
         self._data.erase(parallel=self.config.parallel)
-        del self._data
+        self._data = None
         self._inited_for_start = False
 
     def switch_context(self, new_context: str) -> None:
@@ -683,18 +727,19 @@ class Coverage(TConfigurable):
         if not self._started:                           # pragma: part started
             raise CoverageException("Cannot switch context, coverage is not started")
 
+        assert self._collector is not None
         if self._collector.should_start_context:
             self._warn("Conflicting dynamic contexts", slug="dynamic-conflict", once=True)
 
         self._collector.switch_context(new_context)
 
-    def clear_exclude(self, which: str = 'exclude') -> None:
+    def clear_exclude(self, which: str = "exclude") -> None:
         """Clear the exclude list."""
         self._init()
         setattr(self.config, which + "_list", [])
         self._exclude_regex_stale()
 
-    def exclude(self, regex: str, which: str = 'exclude') -> None:
+    def exclude(self, regex: str, which: str = "exclude") -> None:
         """Exclude source lines from execution consideration.
 
         A number of lists of regular expressions are maintained.  Each list
@@ -711,7 +756,6 @@ class Coverage(TConfigurable):
         """
         self._init()
         excl_list = getattr(self.config, which + "_list")
-        assert isinstance(regex, str)
         excl_list.append(regex)
         self._exclude_regex_stale()
 
@@ -726,7 +770,7 @@ class Coverage(TConfigurable):
             self._exclude_re[which] = join_regex(excl_list)
         return self._exclude_re[which]
 
-    def get_exclude_list(self, which: str = 'exclude') -> List[str]:
+    def get_exclude_list(self, which: str = "exclude") -> list[str]:
         """Return a list of excluded regex strings.
 
         `which` indicates which list is desired.  See :meth:`exclude` for the
@@ -755,9 +799,9 @@ class Coverage(TConfigurable):
 
     def combine(
         self,
-        data_paths: Optional[Iterable[str]] = None,
+        data_paths: Iterable[str] | None = None,
         strict: bool = False,
-        keep: bool = False
+        keep: bool = False,
     ) -> None:
         """Combine together a number of similarly-named coverage data files.
 
@@ -789,6 +833,7 @@ class Coverage(TConfigurable):
         self._post_init()
         self.get_data()
 
+        assert self._data is not None
         combine_parallel_data(
             self._data,
             aliases=self._make_aliases(),
@@ -812,13 +857,15 @@ class Coverage(TConfigurable):
         self._init_data(suffix=None)
         self._post_init()
 
-        for plugin in self._plugins:
-            if not plugin._coverage_enabled:
-                self._collector.plugin_was_disabled(plugin)
+        if self._collector is not None:
+            for plugin in self._plugins:
+                if not plugin._coverage_enabled:
+                    self._collector.plugin_was_disabled(plugin)
 
-        if hasattr(self, "_collector") and self._collector.flush_data():
-            self._post_save_work()
+            if self._collector.flush_data():
+                self._post_save_work()
 
+        assert self._data is not None
         return self._data
 
     def _post_save_work(self) -> None:
@@ -828,6 +875,9 @@ class Coverage(TConfigurable):
         Look for un-executed files.
 
         """
+        assert self._data is not None
+        assert self._inorout is not None
+
         # If there are still entries in the source_pkgs_unmatched list,
         # then we never encountered those packages.
         if self._warn_unimported_source:
@@ -839,16 +889,15 @@ class Coverage(TConfigurable):
 
         # Touch all the files that could have executed, so that we can
         # mark completely un-executed files as 0% covered.
-        if self._data is not None:
-            file_paths = collections.defaultdict(list)
-            for file_path, plugin_name in self._inorout.find_possibly_unexecuted_files():
-                file_path = self._file_mapper(file_path)
-                file_paths[plugin_name].append(file_path)
-            for plugin_name, paths in file_paths.items():
-                self._data.touch_files(paths, plugin_name)
+        file_paths = collections.defaultdict(list)
+        for file_path, plugin_name in self._inorout.find_possibly_unexecuted_files():
+            file_path = self._file_mapper(file_path)
+            file_paths[plugin_name].append(file_path)
+        for plugin_name, paths in file_paths.items():
+            self._data.touch_files(paths, plugin_name)
 
     # Backward compatibility with version 1.
-    def analysis(self, morf: TMorf) -> Tuple[str, List[TLineNo], List[TLineNo], str]:
+    def analysis(self, morf: TMorf) -> tuple[str, list[TLineNo], list[TLineNo], str]:
         """Like `analysis2` but doesn't return excluded line numbers."""
         f, s, _, m, mf = self.analysis2(morf)
         return f, s, m, mf
@@ -856,7 +905,7 @@ class Coverage(TConfigurable):
     def analysis2(
         self,
         morf: TMorf,
-    ) -> Tuple[str, List[TLineNo], List[TLineNo], List[TLineNo], str]:
+    ) -> tuple[str, list[TLineNo], list[TLineNo], list[TLineNo], str]:
         """Analyze a module.
 
         `morf` is a module or a file name.  It will be analyzed to determine
@@ -882,28 +931,22 @@ class Coverage(TConfigurable):
             analysis.missing_formatted(),
         )
 
-    def _analyze(self, it: Union[FileReporter, TMorf]) -> Analysis:
-        """Analyze a single morf or code unit.
-
-        Returns an `Analysis` object.
-
-        """
-        # All reporting comes through here, so do reporting initialization.
+    def _analyze(self, morf: TMorf) -> Analysis:
+        """Analyze a module or file.  Private for now."""
         self._init()
         self._post_init()
 
         data = self.get_data()
-        if isinstance(it, FileReporter):
-            fr = it
-        else:
-            fr = self._get_file_reporter(it)
-
-        return Analysis(data, self.config.precision, fr, self._file_mapper)
+        file_reporter = self._get_file_reporter(morf)
+        filename = self._file_mapper(file_reporter.filename)
+        return analysis_from_file_reporter(data, self.config.precision, file_reporter, filename)
 
+    @functools.lru_cache(maxsize=1)
     def _get_file_reporter(self, morf: TMorf) -> FileReporter:
         """Get a FileReporter for a module or file name."""
+        assert self._data is not None
         plugin = None
-        file_reporter: Union[str, FileReporter] = "python"
+        file_reporter: str | FileReporter = "python"
 
         if isinstance(morf, str):
             mapped_morf = self._file_mapper(morf)
@@ -916,8 +959,8 @@ class Coverage(TConfigurable):
                     if file_reporter is None:
                         raise PluginError(
                             "Plugin {!r} did not provide a file reporter for {!r}.".format(
-                                plugin._coverage_plugin_name, morf
-                            )
+                                plugin._coverage_plugin_name, morf,
+                            ),
                         )
 
         if file_reporter == "python":
@@ -926,17 +969,21 @@ class Coverage(TConfigurable):
         assert isinstance(file_reporter, FileReporter)
         return file_reporter
 
-    def _get_file_reporters(self, morfs: Optional[Iterable[TMorf]] = None) -> List[FileReporter]:
-        """Get a list of FileReporters for a list of modules or file names.
+    def _get_file_reporters(
+        self,
+        morfs: Iterable[TMorf] | None = None,
+    ) -> list[tuple[FileReporter, TMorf]]:
+        """Get FileReporters for a list of modules or file names.
 
         For each module or file name in `morfs`, find a FileReporter.  Return
-        the list of FileReporters.
+        a list pairing FileReporters with the morfs.
 
         If `morfs` is a single module or file name, this returns a list of one
         FileReporter.  If `morfs` is empty or None, then the list of all files
         measured is used to find the FileReporters.
 
         """
+        assert self._data is not None
         if not morfs:
             morfs = self._data.measured_files()
 
@@ -944,30 +991,30 @@ class Coverage(TConfigurable):
         if not isinstance(morfs, (list, tuple, set)):
             morfs = [morfs]     # type: ignore[list-item]
 
-        file_reporters = [self._get_file_reporter(morf) for morf in morfs]
-        return file_reporters
+        return [(self._get_file_reporter(morf), morf) for morf in morfs]
 
     def _prepare_data_for_reporting(self) -> None:
-        """Re-map data before reporting, to get implicit 'combine' behavior."""
+        """Re-map data before reporting, to get implicit "combine" behavior."""
         if self.config.paths:
             mapped_data = CoverageData(warn=self._warn, debug=self._debug, no_disk=True)
-            mapped_data.update(self._data, aliases=self._make_aliases())
+            if self._data is not None:
+                mapped_data.update(self._data, map_path=self._make_aliases().map)
             self._data = mapped_data
 
     def report(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        show_missing: Optional[bool] = None,
-        ignore_errors: Optional[bool] = None,
-        file: Optional[IO[str]] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        skip_covered: Optional[bool] = None,
-        contexts: Optional[List[str]] = None,
-        skip_empty: Optional[bool] = None,
-        precision: Optional[int] = None,
-        sort: Optional[str] = None,
-        output_format: Optional[str] = None,
+        morfs: Iterable[TMorf] | None = None,
+        show_missing: bool | None = None,
+        ignore_errors: bool | None = None,
+        file: IO[str] | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        skip_covered: bool | None = None,
+        contexts: list[str] | None = None,
+        skip_empty: bool | None = None,
+        precision: int | None = None,
+        sort: str | None = None,
+        output_format: str | None = None,
     ) -> float:
         """Write a textual summary report to `file`.
 
@@ -1038,21 +1085,15 @@ class Coverage(TConfigurable):
 
     def annotate(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        directory: Optional[str] = None,
-        ignore_errors: Optional[bool] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        contexts: Optional[List[str]] = None,
+        morfs: Iterable[TMorf] | None = None,
+        directory: str | None = None,
+        ignore_errors: bool | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        contexts: list[str] | None = None,
     ) -> None:
         """Annotate a list of modules.
 
-        .. note::
-
-            This method has been obsoleted by more modern reporting tools,
-            including the :meth:`html_report` method.  It will be removed in a
-            future version.
-
         Each module in `morfs` is annotated.  The source is written to a new
         file, named with a ",cover" suffix, with each line prefixed with a
         marker to indicate the coverage of the line.  Covered lines have ">",
@@ -1061,9 +1102,6 @@ class Coverage(TConfigurable):
         See :meth:`report` for other arguments.
 
         """
-        print("The annotate command will be removed in a future version.")
-        print("Get in touch if you still use it: ned@nedbatchelder.com")
-
         self._prepare_data_for_reporting()
         with override_config(
             self,
@@ -1077,18 +1115,18 @@ class Coverage(TConfigurable):
 
     def html_report(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        directory: Optional[str] = None,
-        ignore_errors: Optional[bool] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        extra_css: Optional[str] = None,
-        title: Optional[str] = None,
-        skip_covered: Optional[bool] = None,
-        show_contexts: Optional[bool] = None,
-        contexts: Optional[List[str]] = None,
-        skip_empty: Optional[bool] = None,
-        precision: Optional[int] = None,
+        morfs: Iterable[TMorf] | None = None,
+        directory: str | None = None,
+        ignore_errors: bool | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        extra_css: str | None = None,
+        title: str | None = None,
+        skip_covered: bool | None = None,
+        show_contexts: bool | None = None,
+        contexts: list[str] | None = None,
+        skip_empty: bool | None = None,
+        precision: int | None = None,
     ) -> float:
         """Generate an HTML report.
 
@@ -1135,13 +1173,13 @@ class Coverage(TConfigurable):
 
     def xml_report(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        outfile: Optional[str] = None,
-        ignore_errors: Optional[bool] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        contexts: Optional[List[str]] = None,
-        skip_empty: Optional[bool] = None,
+        morfs: Iterable[TMorf] | None = None,
+        outfile: str | None = None,
+        ignore_errors: bool | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        contexts: list[str] | None = None,
+        skip_empty: bool | None = None,
     ) -> float:
         """Generate an XML report of coverage results.
 
@@ -1169,14 +1207,14 @@ class Coverage(TConfigurable):
 
     def json_report(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        outfile: Optional[str] = None,
-        ignore_errors: Optional[bool] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        contexts: Optional[List[str]] = None,
-        pretty_print: Optional[bool] = None,
-        show_contexts: Optional[bool] = None,
+        morfs: Iterable[TMorf] | None = None,
+        outfile: str | None = None,
+        ignore_errors: bool | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        contexts: list[str] | None = None,
+        pretty_print: bool | None = None,
+        show_contexts: bool | None = None,
     ) -> float:
         """Generate a JSON report of coverage results.
 
@@ -1207,19 +1245,19 @@ class Coverage(TConfigurable):
 
     def lcov_report(
         self,
-        morfs: Optional[Iterable[TMorf]] = None,
-        outfile: Optional[str] = None,
-        ignore_errors: Optional[bool] = None,
-        omit: Optional[Union[str, List[str]]] = None,
-        include: Optional[Union[str, List[str]]] = None,
-        contexts: Optional[List[str]] = None,
+        morfs: Iterable[TMorf] | None = None,
+        outfile: str | None = None,
+        ignore_errors: bool | None = None,
+        omit: str | list[str] | None = None,
+        include: str | list[str] | None = None,
+        contexts: list[str] | None = None,
     ) -> float:
         """Generate an LCOV report of coverage results.
 
-        Each module in 'morfs' is included in the report. 'outfile' is the
+        Each module in `morfs` is included in the report. `outfile` is the
         path to write the file to, "-" will write to stdout.
 
-        See :meth 'report' for other arguments.
+        See :meth:`report` for other arguments.
 
         .. versionadded:: 6.3
         """
@@ -1234,7 +1272,7 @@ class Coverage(TConfigurable):
         ):
             return render_report(self.config.lcov_output, LcovReporter(self), morfs, self._message)
 
-    def sys_info(self) -> Iterable[Tuple[str, Any]]:
+    def sys_info(self) -> Iterable[tuple[str, Any]]:
         """Return a list of (key, value) pairs showing internal information."""
 
         import coverage as covmod
@@ -1242,7 +1280,7 @@ class Coverage(TConfigurable):
         self._init()
         self._post_init()
 
-        def plugin_info(plugins: List[Any]) -> List[str]:
+        def plugin_info(plugins: list[Any]) -> list[str]:
             """Make an entry for the sys_info from a list of plug-ins."""
             entries = []
             for plugin in plugins:
@@ -1253,41 +1291,35 @@ class Coverage(TConfigurable):
             return entries
 
         info = [
-            ('coverage_version', covmod.__version__),
-            ('coverage_module', covmod.__file__),
-            ('tracer', self._collector.tracer_name() if hasattr(self, "_collector") else "-none-"),
-            ('CTracer', 'available' if HAS_CTRACER else "unavailable"),
-            ('plugins.file_tracers', plugin_info(self._plugins.file_tracers)),
-            ('plugins.configurers', plugin_info(self._plugins.configurers)),
-            ('plugins.context_switchers', plugin_info(self._plugins.context_switchers)),
-            ('configs_attempted', self.config.attempted_config_files),
-            ('configs_read', self.config.config_files_read),
-            ('config_file', self.config.config_file),
-            ('config_contents',
-                repr(self.config._config_contents) if self.config._config_contents else '-none-'
+            ("coverage_version", covmod.__version__),
+            ("coverage_module", covmod.__file__),
+            ("core", self._collector.tracer_name() if self._collector is not None else "-none-"),
+            ("CTracer", "available" if HAS_CTRACER else "unavailable"),
+            ("plugins.file_tracers", plugin_info(self._plugins.file_tracers)),
+            ("plugins.configurers", plugin_info(self._plugins.configurers)),
+            ("plugins.context_switchers", plugin_info(self._plugins.context_switchers)),
+            ("configs_attempted", self.config.config_files_attempted),
+            ("configs_read", self.config.config_files_read),
+            ("config_file", self.config.config_file),
+            ("config_contents",
+                repr(self.config._config_contents) if self.config._config_contents else "-none-",
             ),
-            ('data_file', self._data.data_filename() if hasattr(self, "_data") else "-none-"),
-            ('python', sys.version.replace('\n', '')),
-            ('platform', platform.platform()),
-            ('implementation', platform.python_implementation()),
-            ('executable', sys.executable),
-            ('def_encoding', sys.getdefaultencoding()),
-            ('fs_encoding', sys.getfilesystemencoding()),
-            ('pid', os.getpid()),
-            ('cwd', os.getcwd()),
-            ('path', sys.path),
-            ('environment', human_sorted(
-                f"{k} = {v}"
-                for k, v in os.environ.items()
-                if (
-                    any(slug in k for slug in ("COV", "PY")) or
-                    (k in ("HOME", "TEMP", "TMP"))
-                )
-            )),
-            ('command_line', " ".join(getattr(sys, 'argv', ['-none-']))),
+            ("data_file", self._data.data_filename() if self._data is not None else "-none-"),
+            ("python", sys.version.replace("\n", "")),
+            ("platform", platform.platform()),
+            ("implementation", platform.python_implementation()),
+            ("gil_enabled", getattr(sys, '_is_gil_enabled', lambda: True)()),
+            ("executable", sys.executable),
+            ("def_encoding", sys.getdefaultencoding()),
+            ("fs_encoding", sys.getfilesystemencoding()),
+            ("pid", os.getpid()),
+            ("cwd", os.getcwd()),
+            ("path", sys.path),
+            ("environment", [f"{k} = {v}" for k, v in relevant_environment_display(os.environ)]),
+            ("command_line", " ".join(getattr(sys, "argv", ["-none-"]))),
         ]
 
-        if hasattr(self, "_inorout"):
+        if self._inorout is not None:
             info.extend(self._inorout.sys_info())
 
         info.extend(CoverageData.sys_info())
@@ -1297,16 +1329,16 @@ class Coverage(TConfigurable):
 
 # Mega debugging...
 # $set_env.py: COVERAGE_DEBUG_CALLS - Lots and lots of output about calls to Coverage.
-if int(os.environ.get("COVERAGE_DEBUG_CALLS", 0)):              # pragma: debugging
+if int(os.getenv("COVERAGE_DEBUG_CALLS", 0)):               # pragma: debugging
     from coverage.debug import decorate_methods, show_calls
 
     Coverage = decorate_methods(        # type: ignore[misc]
         show_calls(show_args=True),
-        butnot=['get_data']
+        butnot=["get_data"],
     )(Coverage)
 
 
-def process_startup() -> Optional[Coverage]:
+def process_startup() -> Coverage | None:
     """Call this at Python start-up to perhaps measure coverage.
 
     If the environment variable COVERAGE_PROCESS_START is defined, coverage
@@ -1329,7 +1361,7 @@ def process_startup() -> Optional[Coverage]:
     not started by this call.
 
     """
-    cps = os.environ.get("COVERAGE_PROCESS_START")
+    cps = os.getenv("COVERAGE_PROCESS_START")
     if not cps:
         # No request for coverage, nothing to do.
         return None
