@@ -8,11 +8,11 @@ from __future__ import annotations
 import contextlib
 import datetime
 import errno
+import functools
 import hashlib
 import importlib
 import importlib.util
 import inspect
-import locale
 import os
 import os.path
 import re
@@ -21,11 +21,9 @@ import types
 
 from types import ModuleType
 from typing import (
-    Any, Callable, Dict, IO, Iterable, Iterator, List, Mapping, Optional,
-    Sequence, Tuple, TypeVar, Union,
+    Any, Iterable, Iterator, Mapping, NoReturn, Sequence, TypeVar,
 )
 
-from coverage import env
 from coverage.exceptions import CoverageException
 from coverage.types import TArc
 
@@ -34,7 +32,7 @@ from coverage.types import TArc
 # pylint: disable=unused-wildcard-import
 from coverage.exceptions import *   # pylint: disable=wildcard-import
 
-ISOLATED_MODULES: Dict[ModuleType, ModuleType] = {}
+ISOLATED_MODULES: dict[ModuleType, ModuleType] = {}
 
 
 def isolate_module(mod: ModuleType) -> ModuleType:
@@ -80,7 +78,7 @@ def sys_modules_saved() -> Iterator[None]:
         saver.restore()
 
 
-def import_third_party(modname: str) -> Tuple[ModuleType, bool]:
+def import_third_party(modname: str) -> tuple[ModuleType, bool]:
     """Import a third-party module we need, but might not be installed.
 
     This also cleans out the module after the import, so that coverage won't
@@ -117,30 +115,7 @@ def nice_pair(pair: TArc) -> str:
         return "%d-%d" % (start, end)
 
 
-TSelf = TypeVar("TSelf")
-TRetVal = TypeVar("TRetVal")
-
-def expensive(fn: Callable[[TSelf], TRetVal]) -> Callable[[TSelf], TRetVal]:
-    """A decorator to indicate that a method shouldn't be called more than once.
-
-    Normally, this does nothing.  During testing, this raises an exception if
-    called more than once.
-
-    """
-    if env.TESTING:
-        attr = "_once_" + fn.__name__
-
-        def _wrapper(self: TSelf) -> TRetVal:
-            if hasattr(self, attr):
-                raise AssertionError(f"Shouldn't have called {fn.__name__} more than once")
-            setattr(self, attr, True)
-            return fn(self)
-        return _wrapper
-    else:
-        return fn                   # pragma: not testing
-
-
-def bool_or_none(b: Any) -> Optional[bool]:
+def bool_or_none(b: Any) -> bool | None:
     """Return bool(b), but preserve None."""
     if b is None:
         return None
@@ -180,18 +155,6 @@ def ensure_dir_for_file(path: str) -> None:
     ensure_dir(os.path.dirname(path))
 
 
-def output_encoding(outfile: Optional[IO[str]] = None) -> str:
-    """Determine the encoding to use for output written to `outfile` or stdout."""
-    if outfile is None:
-        outfile = sys.stdout
-    encoding = (
-        getattr(outfile, "encoding", None) or
-        getattr(sys.__stdout__, "encoding", None) or
-        locale.getpreferredencoding()
-    )
-    return encoding
-
-
 class Hasher:
     """Hashes Python data for fingerprinting."""
     def __init__(self) -> None:
@@ -218,21 +181,21 @@ class Hasher:
                 self.update(v[k])
         else:
             for k in dir(v):
-                if k.startswith('__'):
+                if k.startswith("__"):
                     continue
                 a = getattr(v, k)
                 if inspect.isroutine(a):
                     continue
                 self.update(k)
                 self.update(a)
-        self.hash.update(b'.')
+        self.hash.update(b".")
 
     def hexdigest(self) -> str:
         """Retrieve the hex digest of the hash."""
         return self.hash.hexdigest()[:32]
 
 
-def _needs_to_implement(that: Any, func_name: str) -> None:
+def _needs_to_implement(that: Any, func_name: str) -> NoReturn:
     """Helper to raise NotImplementedError in interface stubs."""
     if hasattr(that, "_coverage_plugin_name"):
         thing = "Plugin"
@@ -243,7 +206,7 @@ def _needs_to_implement(that: Any, func_name: str) -> None:
         name = f"{klass.__module__}.{klass.__name__}"
 
     raise NotImplementedError(
-        f"{thing} {name!r} needs to implement {func_name}()"
+        f"{thing} {name!r} needs to implement {func_name}()",
     )
 
 
@@ -292,7 +255,7 @@ def substitute_variables(text: str, variables: Mapping[str, str]) -> str:
         )
         """
 
-    dollar_groups = ('dollar', 'word1', 'word2')
+    dollar_groups = ("dollar", "word1", "word2")
 
     def dollar_replace(match: re.Match[str]) -> str:
         """Called for each $replacement."""
@@ -302,11 +265,11 @@ def substitute_variables(text: str, variables: Mapping[str, str]) -> str:
             return "$"
         elif word in variables:
             return variables[word]
-        elif match['strict']:
+        elif match["strict"]:
             msg = f"Variable {word} is undefined: {text!r}"
             raise CoverageException(msg)
         else:
-            return match['defval']
+            return match["defval"]
 
     text = re.sub(dollar_pattern, dollar_replace, text)
     return text
@@ -315,10 +278,10 @@ def substitute_variables(text: str, variables: Mapping[str, str]) -> str:
 def format_local_datetime(dt: datetime.datetime) -> str:
     """Return a string with local timezone representing the date.
     """
-    return dt.astimezone().strftime('%Y-%m-%d %H:%M %z')
+    return dt.astimezone().strftime("%Y-%m-%d %H:%M %z")
 
 
-def import_local_file(modname: str, modfile: Optional[str] = None) -> ModuleType:
+def import_local_file(modname: str, modfile: str | None = None) -> ModuleType:
     """Import a local file as a module.
 
     Opens a file in the current directory named `modname`.py, imports it
@@ -327,7 +290,7 @@ def import_local_file(modname: str, modfile: Optional[str] = None) -> ModuleType
 
     """
     if modfile is None:
-        modfile = modname + '.py'
+        modfile = modname + ".py"
     spec = importlib.util.spec_from_file_location(modname, modfile)
     assert spec is not None
     mod = importlib.util.module_from_spec(spec)
@@ -338,20 +301,25 @@ def import_local_file(modname: str, modfile: Optional[str] = None) -> ModuleType
     return mod
 
 
-def _human_key(s: str) -> List[Union[str, int]]:
+@functools.lru_cache(maxsize=None)
+def _human_key(s: str) -> tuple[list[str | int], str]:
     """Turn a string into a list of string and number chunks.
-        "z23a" -> ["z", 23, "a"]
+
+    "z23a" -> (["z", 23, "a"], "z23a")
+
+    The original string is appended as a last value to ensure the
+    key is unique enough so that "x1y" and "x001y" can be distinguished.
     """
-    def tryint(s: str) -> Union[str, int]:
+    def tryint(s: str) -> str | int:
         """If `s` is a number, return an int, else `s` unchanged."""
         try:
             return int(s)
         except ValueError:
             return s
 
-    return [tryint(c) for c in re.split(r"(\d+)", s)]
+    return ([tryint(c) for c in re.split(r"(\d+)", s)], s)
 
-def human_sorted(strings: Iterable[str]) -> List[str]:
+def human_sorted(strings: Iterable[str]) -> list[str]:
     """Sort the given iterable of strings the way that humans expect.
 
     Numeric components in the strings are sorted as numbers.
@@ -366,7 +334,7 @@ SortableItem = TypeVar("SortableItem", bound=Sequence[Any])
 def human_sorted_items(
     items: Iterable[SortableItem],
     reverse: bool = False,
-) -> List[SortableItem]:
+) -> list[SortableItem]:
     """Sort (string, ...) items the way humans expect.
 
     The elements of `items` can be any tuple/list. They'll be sorted by the
@@ -386,3 +354,15 @@ def plural(n: int, thing: str = "", things: str = "") -> str:
         return thing
     else:
         return things or (thing + "s")
+
+
+def stdout_link(text: str, url: str) -> str:
+    """Format text+url as a clickable link for stdout.
+
+    If attached to a terminal, use escape sequences. Otherwise, just return
+    the text.
+    """
+    if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
+        return f"\033]8;;{url}\a{text}\033]8;;\a"
+    else:
+        return text
